#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import contextlib

from xotl.tools.context import context


def ReentrantContext(context_identifier):
    """Return a reentrant context manager"""

    @contextlib.contextmanager
    def reentrant_context():
        ctx = context[context_identifier]
        if ctx:
            yield ctx["memory"]
        else:
            with context(context_identifier, memory={}) as ctx:
                yield ctx["memory"]

    return reentrant_context



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import contextlib
import sys


@contextlib.contextmanager
def captured_output(stream_name):
    """Return a context manager used by captured_stdout/stdin/stderr
    that temporarily replaces the sys stream *stream_name* with a StringIO."""
    import io

    orig_stdout = getattr(sys, stream_name)
    setattr(sys, stream_name, io.StringIO())
    try:
        yield getattr(sys, stream_name)
    finally:
        setattr(sys, stream_name, orig_stdout)


def captured_stdout():
    """Capture the output of sys.stdout:

    with captured_stdout() as stdout:
        print("hello")
    self.assertEqual(stdout.getvalue(), "hello\\n")
    """
    return captured_output("stdout")


def captured_stderr():
    """Capture the output of sys.stderr:

    with captured_stderr() as stderr:
        print("hello", file=sys.stderr)
    self.assertEqual(stderr.getvalue(), "hello\\n")
    """
    return captured_output("stderr")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "xotless"
author = "Merchise Autrement"

from datetime import datetime  # noqa

# Any year before to 2012 xoutil copyrights to "Medardo Rodriguez"
copyright = f"2012-{datetime.now().year} {author} [~º/~] and Contributors"
del datetime


# The full version, including alpha/beta/rc tags
try:
    from xotless.release import VERSION
except ImportError:

    def up(path, level=1):
        result = path
        while level:
            result = os.path.dirname(result)
            level -= 1
        return result

    _current_dir = os.path.dirname(os.path.abspath(__file__))
    _project_dir = os.path.abspath(os.path.join(up(_current_dir, 2)))
    sys.path.append(_project_dir)
    from xotless.release import VERSION
version = VERSION[: VERSION.rfind(".")]

try:
    from xotless.release import RELEASE_TAG

    release = VERSION
    if RELEASE_TAG:
        release += RELEASE_TAG
except:
    # The full version, including alpha/beta/rc tags.
    release = VERSION


# -- General configuration ---------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
try:
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except ImportError:
    html_theme = "pyramid"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "py3": ("https://docs.python.org/3.8/", None),
    "hypothesis": ("https://hypothesis.readthedocs.io/en/latest/", None),
    "xotl.tools": ("https://xoutil.readthedocs.io/en/latest/", None),
}

# Maintain the cache forever.
intersphinx_cache_limit = 365

autosummary_generate = True

# Produce output for todo and todolist
todo_include_todos = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# copyright (c) merchise autrement [~º/~] and contributors
# all rights reserved.
#
# this is free software; you can do what the licence file allows you to.
#
"""Implements trees for efficient searching.


"""
import operator
from dataclasses import dataclass
from itertools import takewhile
from typing import Generic, Iterator, Optional, Sequence, Tuple, TypeVar

from .ranges import Range, TOrd

T = TypeVar("T")


@dataclass(unsafe_hash=True)
class Cell(Generic[T, TOrd]):
    """A data-carrying open-right range."""

    # NB: Introducing any kind of Range here entails modifications to the
    # algorithms below.
    lowerbound: TOrd
    upperbound: TOrd
    data: T

    __slots__ = ("lowerbound", "upperbound", "data")

    @classmethod
    def from_bounds(
        cls, lower: TOrd, upper: TOrd, data: T
    ) -> "Cell[T, TOrd]":  # pragma: no cover
        """Return a cell from the boundaries of an open-right range."""
        return cls(lower, upper, data)

    @classmethod
    def from_range(cls, r: Range[TOrd], data: T) -> "Cell[T, TOrd]":
        """Return a cell casting `r` to an open-right range.

        .. note:: We disregard the kind of range, cells are always open-right.

        """
        return cls(r.lowerbound, r.upperbound, data)

    def __contains__(self, which):  # pragma: no cover
        return self.lowerbound <= which < self.upperbound

    def __bool__(self) -> bool:  # pragma: no cover
        return self.lowerbound < self.upperbound


@dataclass
class IntervalTree(Generic[T, TOrd]):
    """A data-containing generic IntervalTree.

    See section 10.1 of [deBerg2008]_.

    The intervals are allowed over any type `TOrd` (i.e, for which `<=` is
    defined and total).

    The `cells` can contain any type of data.

    You shouldn't mutate any of the internal attributes.  The API is just the
    class method `from_cells` and then getting the cells containing a point
    the with either `get` of `__getitem__`.

    .. [deBerg2008] Mark de Berg, et al. *Computational Geometry. Algorithms
       and Applications. Third Edition*. Springer Verlag. 2008.

    """

    center: Optional[TOrd]
    cells: Sequence[Cell[T, TOrd]]  # Ordered by lowerbound
    cells_by_upper: Sequence[Cell[T, TOrd]]  # Ordered in reverse by upperbound.
    left: Optional["IntervalTree"]
    right: Optional["IntervalTree"]

    __slots__ = ("center", "cells", "cells_by_upper", "left", "right")

    @classmethod
    def from_cells(cls, cells: Sequence[Cell[T, TOrd]]) -> "IntervalTree":
        return cls._from_sorted_cells(
            sorted(tuple(cell for cell in cells if cell), key=_getlower)
        )

    def __bool__(self) -> bool:
        return bool(self.cells)

    def __contains__(self, which: TOrd):
        return any(which in cell for cell in self.cells)

    @classmethod
    def _from_sorted_cells(
        cls, cells: Sequence[Cell[T, TOrd]]
    ) -> "IntervalTree[T, TOrd]":
        left_cells = []
        middle_cells = []
        right_cells = []
        if cells:
            middle = len(cells) // 2
            middle_cell = cells[middle]
            center: Optional[TOrd] = middle_cell.lowerbound
            for cell in cells:
                if cell.upperbound <= center:
                    left_cells.append(cell)
                elif center < cell.lowerbound:
                    right_cells.append(cell)
                else:
                    assert center in cell
                    middle_cells.append(cell)
        else:
            center = None
        result = cls(
            center=center,
            cells=tuple(middle_cells),
            cells_by_upper=tuple(
                sorted(middle_cells, key=_getupper, reverse=True)
            ),
            left=cls._from_sorted_cells(left_cells) if left_cells else None,
            right=cls._from_sorted_cells(right_cells) if right_cells else None,
        )
        return result

    def __getitem__(self, which: TOrd) -> Tuple[Cell[T, TOrd], ...]:
        """Return all cells that contains the given item."""
        result = self.get(which)
        if not result:
            raise KeyError(which)
        return result

    def get(self, query: TOrd) -> Tuple[Cell[T, TOrd], ...]:
        return tuple(self.iter_get(query))

    def iter_get(self, query: TOrd) -> Iterator[Cell[T, TOrd]]:
        if not self:
            return
        if query <= self.center:
            # Since self.cells is ordered by lowerbound, only *the first*
            # cells for which the lowerbound is less or equal to the query
            # contain the queried value.  Looking any further won't find any
            # match.
            #
            # See page 222 of [deBerg2008].
            yield from takewhile(lambda c: c.lowerbound <= query, self.cells)
        else:
            # The same as before but with `cells_by_upper`.
            yield from takewhile(
                lambda c: query < c.upperbound, self.cells_by_upper
            )
        if query < self.center and self.left:
            yield from self.left.get(query)
        elif query > self.center and self.right:
            yield from self.right.get(query)


_getlower = operator.attrgetter("lowerbound")
_getupper = operator.attrgetter("upperbound")
